"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var _1 = require("../");
_1.setMeta('/clause', {
    'name': 'C',
    'showAsFunction': true,
    'comment': 'Convenient method that combines C.get() and C.set() into one function to get or set clauses from the global registry.',
    'examples': [
        'C("awesomeapp/TodoItem", TodoItemClause)',
        "var ref = C(\"awesomeapp/TodoItem\");\nref.get()"
    ],
    'args': {
        'register': {
            'comment': 'Put the given expression onto the given path in the global clause registry.',
            'examples': 'C("awesomeapp/TodoItem", TodoItemClause)',
        },
        'retrieve': {
            'comment': 'Retrieves an expression from the given namespace path, or returns null if not found.',
            'examples': 'C("awesomeapp/TodoItem")',
        }
    },
    'ret': {
        'comment': 'Note: The returned Clause is of type ClauseRef and is not immediately resolved until one calls its get() method.',
    },
});
_1.setMeta('clause.compose.regex/cat', {
    comment: 'Given a set of expressions or a set of label+expression pairs, returns a regex op that matches all values in an iterable. If labels are given, returns a map that groups values by these labels. Returns a Problem if no match is found.',
    examples: [
        "var CatClause = C.cat(C.isStr, C.isNum, C.isBool);\nC.isValid( CatClause, [\"a\", 2, true] )",
        'C.isValid( CatClause, ["a", false, "b"] )',
        "var LabelledCatClause = C.cat(\n  'first', C.isStr, \n  'second', C.isNum, \n  'third', C.isBool\n);\nC.conform( LabelledCatClause, [\"a\", 2, true] )",
        "// equivalent to the above\nvar LabelledCatClause = C.cat([\n  [ 'first', C.isStr ], \n  [ 'second', 'this is an optional comment on second', C.isNum ], \n  [ 'third', C.isBool ]\n]);\nC.conform( LabelledCatClause, [\"a\", 2, true] )",
    ]
});
_1.setMeta('clause.compose.regex/or', {
    comment: 'Given a set of expressions or a set of label+expression pairs, returns a regex op that matches a value with either one of the expressions. If labels are given, returns a map with a single key that indicates the matching label.',
    examples: [
        "var OrClause = C.or(C.isStr, C.isNum);\nC.conform( OrClause, \"a\" )",
        "var LabelledOrClause = C.or('option1', C.isStr, 'option2', C.isNum );\nC.conform( LabelledOrClause, 2 )",
        "// equivalent to the above\nvar LabelledOrClause = C.or([['option1', C.isStr], ['option2', C.isNum]] );\nC.conform( LabelledOrClause, 2 )",
    ]
});
_1.setMeta('clause.compose.regex/zeroOrMore', {
    comment: 'Given a single expression, returns a regex op that matches zero or more values in an iterable. Returns a Problem if there are no matches.',
    examples: [
        "var ZOMClause = C.zeroOrMore(C.isNum);\nC.isValid( ZOMClause, [1, 2, 3] )",
        'C.isValid( ZOMClause, [1, 2, "3"] )',
        'C.isValid( ZOMClause, [] )',
        "var CombinedClause = C.cat( \n  \"babbles\", C.zeroOrMore(C.oneOf(\"foo\", \"bar\", \"baz\")),\n  \"truths\", C.zeroOrMore(C.isBool),\n  \"counts\", C.zeroOrMore(C.isNatInt)\n);\nC.conform( CombinedClause, [\"foo\", \"foo\", \"bar\", true,  2, 3 , 4] )",
        'C.isValid( CombinedClause, ["foo", "foo", "bar", 2, 3 , 4] )',
        'C.isValid( CombinedClause, [ 2, 3 , 4 ] )',
        'C.isValid( CombinedClause, [ ] )',
    ]
});
_1.setMeta('clause.compose.regex/oneOrMore', {
    comment: 'Given a single expression, returns a regex op that matches one or more values in an iterable. Returns a Problem if there are no matches.',
    examples: [
        "var OOMClause = C.oneOrMore(C.equals(\"hello\"));\nC.isValid( OOMClause, [ \"hello\", \"hello\", \"hello\" ] )",
        'C.isValid( OOMClause, [ "hello", "hello", "hello" ] )',
        "var CombinedClause = C.cat( \n  \"babbles\", C.oneOrMore(C.oneOf(\"foo\", \"bar\", \"baz\")),\n  \"truths\", C.oneOrMore(C.isBool),\n  \"counts\", C.oneOrMore(C.isNatInt)\n);\nC.conform( CombinedClause, [\"foo\", \"foo\", \"baz\", false, true, 2, 3 ] )",
        'C.isValid( CombinedClause, [false, true, 2, 3 ] )',
    ]
});
_1.setMeta('clause.compose.regex/zeroOrOne', {
    comment: 'Given a single expression, returns a regex op that matches zero or one values in an iterable. Returns a Problem if there are no matches.',
    examples: [
        "var ZOOClause = C.zeroOrOne( C.equals( \"hello\" ) );\nC.isValid( ZOOClause, [ \"hello\" ] )",
        'C.isValid( ZOOClause, [] )',
        "var CombinedClause = C.cat( \n  \"the babble\", C.zeroOrOne( C.oneOf(\"foo\", \"bar\", \"baz\") ),\n  \"the truth\", C.zeroOrOne( C.isBool ),\n  \"the count\", C.zeroOrOne( C.isNatInt )\n);\nC.conform( CombinedClause, [\"foo\", 4] )",
    ]
});
_1.setMeta('clause.compose/and', {
    comment: 'Given a list of expressions, returns a clause that matches a value with all the expressions. Successive conformed value propagates to the rest of the expressions.',
    examples: [
        "var AndClause = C.and( C.isStr, (s) => s.length > 5 );\nC.isValid( AndClause, \"abcdefgh\" )",
        'C.isValid( AndClause, "abc" )',
        "var PropagatingAndClause = C.and( \n  C.cat( \n    \"first\", C.isStr, \n    \"second\", C.zeroOrMore( C.isNum )\n  ),\n  // \"second\" propagated from conformed results of the previous clause\n  ({ second }) => second.length > 4 ) \nC.isValid( PropagatingAndClause, [ \"foo\", 2, 3, 4, 5, 6 ] )",
        'C.conform( PropagatingAndClause, [ "foo", 2, 3, 4, 5, 6 ] )',
        'C.isValid( PropagatingAndClause, [ "foo", 2, 3 ] )',
    ]
});
_1.setMeta('clause.compose/collOf', {
    comment: 'Given a single expression, returns a clause that matches all values in an iterable with the given expression. ',
    examples: [
        "var CollOfStrClause = C.collOf(C.isStr);\nC.conform( CollOfStrClause, [\"a\", \"b\", \"c\"] )",
        'C.conform( CollOfStrClause, [] )',
        " // notice the difference in behavior from C.zeroOrOne\nvar CombinedCollOfClause = C.cat( \n  \"babbles\", C.collOf(C.oneOf(\"foo\", \"bar\", \"baz\")),\n  \"truths\", C.collOf(C.isBool),\n  \"counts\", C.collOf(C.isNatInt)\n);\nC.isValid( CombinedCollOfClause, [[\"foo\", \"foo\", \"bar\"], [], [2, 3 , 4]] )",
        "// invalid because collOf does not mix with regex ops like zeroOrMore\n// collOf(...) here is equivalent to wall(zeroOrMore(...))\nC.isValid( CombinedCollOfClause, [\"foo\", \"foo\", \"bar\", 2, 3 , 4] )"
    ]
});
_1.setMeta('clause.compose/mapOf', {
    comment: 'Given a pair of expressions, returns a clause that matches an object all of whose keys matches the first expression and all of whose values matches the second expression.',
    examples: [
        "var AbilityMapClause = C.mapOf(\n      (key) => key.indexOf(\"can\") === 0,\n      C.isBool );\nC.isValid( AbilityMapClause, { canFly: true, canSwim: true, canDance: false } )",
        'C.isValid( AbilityMapClause, { canFly: true, canSwim: true, canDance: "huh?" } )',
        'C.isValid( AbilityMapClause, { meNotValid: true, canSwim: true } )',
    ]
});
_1.setMeta('clause.compose/shape', {
    comment: 'Given an object with two optional properties "required" and "optional", returns a clause that matches an object all of whose keys matches the expressions defined in "required" object and some of whose values matches the expressions defined in "optional" objects.',
    examples: [
        "var ShapeClause = C.shape( { \n  required: [\"key1\", \"key2\", \"key3\"], \n  optional: [\"key4\"],\n  } );\nC.isValid( ShapeClause, { key1: true, key2: 2, key3: \"ss\", key4: false } )",
        'C.isValid( ShapeClause, { key1: true, key2: 2, key3: "ss" } )',
        'C.isValid( ShapeClause, { key1: true, key2: 2 } )',
        "var ShapeClause2 = C.shape( { \n  required: {\n    key1: C.isBool,\n    key2: C.isNatInt,\n  }, \n  optional: [\"key5\"],\n  } );\nC.isValid( ShapeClause2, { key1: true, key2: 2, key5: false } )",
        'C.isValid( ShapeClause2, { key1: "not bool", key2: 2, key5: false } )',
        "// key group conformation\nvar ShapeClause3 = C.shape( { \n  required: {\n    abilityGroup: [ \n      // key expression\n      (key) => key.indexOf('can') === 0,\n      // value expression\n      C.isBool\n    ],\n    someKey: C.isNatInt,\n  }, \n  optional: [\"key5\"],\n  } );\nC.conform( ShapeClause3, { \n  canDance: true, \n  canFly: true, \n  canSwim: false, \n  someKey: 999 \n} )",
    ]
});
_1.setMeta('clause.compose/any', {
    comment: 'Returns a clause that matches any value.',
    examples: [
        'C.isValid(C.any(), 123)',
        'C.isValid(C.any(), true)',
        'C.isValid(C.any(), ["sass"])',
        'C.isValid(C.any(), null)',
        'C.isValid(C.any(), undefined)',
    ]
});
_1.setMeta('clause.compose/wall', {
    comment: 'Given a single expression, returns a clause that does not regex-mix with the parent clause. Useful for validating a list inside another list in the sequence.',
    examples: [
        "var FreeMixClause = C.cat( C.oneOrMore( C.isNum ) );\nC.isValid( FreeMixClause, [2, 3, 4] )",
        'C.isValid( FreeMixClause, [ [2, 3, 4] ] )',
        " var WalledClause = C.cat( C.wall( C.oneOrMore( C.isNum ) ) );;\nC.isValid( WalledClause, [2, 3, 4] )",
        'C.isValid( WalledClause, [ [2, 3, 4] ] )',
    ]
});
_1.setMeta('clause.compose/fclause', {
    comment: 'Given an object that contains "args", "ret" and "fn" expressions (all optionally), returns a clause whose instrument function takes a function and validates its arguments and return value during run time and whose instrumentConformed function takes a function that takes conformed arguments.',
    examples: [
        "// first, we define a cat clause\nvar FnArgsClause = C.zeroOrMore(\n   C.cat(\n     'name', C.isStr,\n     'age', C.isNatInt,\n     // optional field\n     'isMember', C.zeroOrOne( C.isBool )\n   )\n );\nC.conform(FnArgsClause, [\"john\", 23, true, \"mary\", 25, \"bob\", 34])",
        "// next, we define a fclause with args as the cat clause defined above\nvar MyFnSpec = C.fclause( {\n args: FnArgsClause,\n ret: C.isNatInt\n} );\n\nvar nameCount = MyFnSpec.instrument( \n  function _nameCount() { return Array.from(arguments).filter((s) => C.isStr(s)).length; } \n);\nnameCount(\"john\", 23, true, \"mary\", 25, \"bob\", 34);\n",
        "// invalid: missing name\n    nameCount(\"john\", 23, true, 24, false);",
        "var averageAge = MyFnSpec.instrumentConformed( \n  function (cprofiles) { // \"c\" stands for \"conformed\"\n    // cprofiles example:\n    // [ { name: \"john\", age: 4, isMember: true }, { name: \"mary\", age: 23 }, { name: \"bob\", age: 102 } ]\n    var sumOfAges = cprofiles\n      .map( ( { age } ) => age )\n      .reduce( (acc, c) => c + acc, 0 );\n    return sumOfAges / cprofiles.length;\n  }\n);\n\n// cprofile below conforms into \naverageAge(\"john\", 4, true, \"mary\", 23, \"bob\", 102);"
    ]
});
_1.setMeta('clause.compose/delayed', {
    comment: 'Given a funciton that returns a clause, returns a DelayedClause that does not evaluate until its get() function is called. Useful for constructing recursive clause definitions.',
    examples: [
        "\nvar BinaryTreeClause = C.maybe( C.wall(\n  C.cat(\n    'nodeName', C.isStr, \n    'leftNode', C.delayed( () => BinaryTreeClause ),\n    'rightNode', C.delayed( () => BinaryTreeClause )\n  )\n) );\n\nC.isValid(\n  BinaryTreeClause, \n  ['john', ['bob', ['alice', null, null], null], ['sam', null, ['jane', null, null]]]\n); ",
    ]
});
_1.setMeta('clause.compose/nullable', {
    comment: 'Given an expression, returns a new clause that matches either the original clause or a null value.',
    examples: [
        'C.isValid(C.isStr, null);',
        'C.isValid(C.nullable(C.isStr), null);'
    ]
});
_1.setMeta('clause.compose/undefinable', {
    comment: 'Given an expression, returns a new clause that matches either the original clause or undefined.',
    examples: [
        'C.isValid(C.isStr, undefined);',
        'C.isValid(C.undefinable(C.isStr), undefined);'
    ]
});
_1.setMeta('clause.compose/maybe', {
    comment: 'Given an expression, returns a new clause that matches either the original clause, a null value, or undefined.',
    examples: [
        'C.isValid(C.isObj, undefined);',
        'C.isValid(C.maybe(C.isObj), null);',
        'C.isValid(C.maybe(C.isObj), undefined);'
    ]
});
_1.setMeta('clause.namespace/get', {
    comment: 'Gets a clause reference from global clause registry.',
    examples: [
        'C("awesomeapp/TodoItem")',
        'C("awesomeapp/TodoItem").get()'
    ]
});
_1.setMeta('clause.namespace/set', {
    'comment': 'Registers a clause in to global clause registry.',
    examples: ['C("awesomeapp/TodoItem", TodoItemClause)']
});
_1.setMeta('clause.compose.string/scat', {
    examples: "\nvar StrClause = C.cat('greetings', C.zeroOrMore(C.scat('hello')),\n                      'substence', C.zeroOrMore(C.scat('i am optional')),\n                      'farewell', C.zeroOrMore(C.scat('bye')));\n\nC.conform(StrClause, 'hellohellobyebyebyebye');\n"
});
_1.setMeta('clause.types/NamespacePath', {
    'comment': 'Represents a namespace path.',
    'examples': '"com.xyz.awesomeApp/User"',
});
_1.setMeta('clause.types/Expression', {
    'comment': 'Represents an Clause expression.',
    'examples': 'isPositiveNumber(x); C.cat(...)',
});
_1.setMeta('clause.utils/describe', {
    'comment': 'Returns an abbreviated description of the clause as a simple tree structure.',
});
//# sourceMappingURL=index.annotation.js.map